from .instances import get_simple_instance_headroom_target
from .instances import get_simple_instance_headroom_target_for_name

__all__ = [
    "get_simple_instance_headroom_target",
    "get_simple_instance_headroom_target_for_name",
]
